# a filterable combobox widget
# layout: text, combobox, filterbuttons
# filterbuttons open a menu with filter checkboxes

import sys
from PyQt5.QtWidgets import (
    QApplication, QWidget, QLabel, QComboBox, QPushButton, QHBoxLayout, QVBoxLayout,
    QDialog, QCheckBox, QDialogButtonBox, QFormLayout, QMenu, QAction, QWidgetAction
)
from PyQt5.QtCore import Qt, QSortFilterProxyModel, QAbstractListModel, QModelIndex, QVariant, QPoint, pyqtSignal
from koalafolio.gui.widgets.QCompleterComboBox import QCompleterComboBoxView

class QStringPropertyListModel(QAbstractListModel):
    def __init__(self, string_list=[], properties_list=[], parent=None):
        super().__init__(parent)
        self._strings = string_list
        self._properties = properties_list

    def setDataFromDict(self, dataDict, string_key="name", properties_keys=["type"]):
        self.beginResetModel()
        self._strings = [item[string_key] for item in dataDict]
        self._properties = [{key: item.get(key, None) for key in properties_keys} for item in dataDict]
        self.endResetModel()

    def rowCount(self, parent=QModelIndex()):
        return len(self._strings)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()
        row = index.row()
        if role == Qt.DisplayRole:
            return self._strings[row]
        if role == Qt.UserRole:
            return self._properties[row]
        return QVariant()


class FilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.active_filters = {}  # key: set of allowed values

    def set_filter(self, key, allowed_values):
        self.active_filters[key] = set(allowed_values)
        self.invalidateFilter()

    def clear_filter(self, key):
        if key in self.active_filters:
            del self.active_filters[key]
            self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        model = self.sourceModel()
        props = model.data(model.index(source_row, 0), Qt.UserRole)
        for key, allowed in self.active_filters.items():
            if allowed and props.get(key) not in allowed:
                return False
        return True


class QFilterableComboBoxView(QWidget):
    currentTextChanged = pyqtSignal([str])
    currentIndexChanged = pyqtSignal([int])
    activated = pyqtSignal([int])
    editTextChanged = pyqtSignal([str])
    highlighted = pyqtSignal([int])

    def __init__(self, label_text="", parent=None):
        super().__init__(parent)
        self.proxyModel = FilterProxyModel(self)
        self.filter_properties = {}  # key: list of possible values
        self.active_filter_values = {}  # key: set of currently enabled values

        self.label = QLabel(label_text, self)
        self.combobox = QCompleterComboBoxView(self)
        # map signals from QComboBox to parent widget
        self.combobox.currentTextChanged.connect(self.currentTextChanged)
        self.combobox.currentIndexChanged.connect(self.currentIndexChanged)
        self.combobox.activated.connect(self.activated)
        self.combobox.editTextChanged.connect(self.editTextChanged)
        self.combobox.highlighted.connect(self.highlighted)

        self.combobox.setModel(self.proxyModel)

        self.filter_buttons = {}  # key: QPushButton
        self.filter_menus = {}  # key: QMenu

        layout = QHBoxLayout(self)
        layout.addWidget(self.label)
        self.filter_buttons_layout = QHBoxLayout()
        layout.addLayout(self.filter_buttons_layout)
        # add combobox with maximum span
        layout.addWidget(self.combobox, 1)
        self.setLayout(layout)

    def setModel(self, model):
        self.proxyModel.setSourceModel(model)
        self.generateFilterProperties(model)
        self.createFilterButtons()
        self.createFilterMenus()
        model.modelReset.connect(self.onModelReset)

    def onModelReset(self):
        self.generateFilterProperties(self.proxyModel.sourceModel())
        self.createFilterButtons()
        self.createFilterMenus()

    def currentText(self):
        return self.combobox.currentText()

    def setCurrentText(self, text):
        self.combobox.setCurrentText(text)

    def setCurrentIndex(self, index):
        self.combobox.setCurrentIndex(index)
        self.combobox.lineEdit().setText(self.proxyModel.data(self.proxyModel.index(index, 0), Qt.DisplayRole))

    def findText(self, text):
        return self.combobox.findText(text)

    def clear(self):
        self.combobox.clear()

    def generateFilterProperties(self, model):
        # Only accept key: list properties
        self.filter_properties.clear()
        for row in range(model.rowCount()):
            props = model.data(model.index(row, 0), Qt.UserRole)
            for key, value in props.items():
                if key not in self.filter_properties:
                    self.filter_properties[key] = []
                if value not in self.filter_properties[key]:
                    self.filter_properties[key].append(value)
        # At start, all values are enabled
        self.active_filter_values = {k: set(v) for k, v in self.filter_properties.items()}

    def createFilterButtons(self):
        # Remove old buttons
        for btn in self.filter_buttons.values():
            self.filter_buttons_layout.removeWidget(btn)
            btn.deleteLater()
        self.filter_buttons.clear()
        # Create new buttons
        for key, values in self.filter_properties.items():
            btn = QPushButton(key, self)
            btn.setCheckable(False)
            btn.clicked.connect(lambda checked, k=key: self.open_filter_popup(k))
            self.filter_buttons_layout.addWidget(btn)
            self.filter_buttons[key] = btn
            
    def createFilterMenus(self):
        self.filter_menus = {}
        for key, values in self.filter_properties.items():
            self.filter_menus[key] = self._create_single_filter_menu(key, values)

    def _create_single_filter_menu(self, key, values):
        menu = QMenu(self)
        menu.setSeparatorsCollapsible(False)
        actions = {}
        max_width = 0

        # Add Select All checkbox at the top using QWidgetAction
        select_all_widget = QCheckBox("Select All", menu)
        select_all_action = QWidgetAction(menu)
        select_all_action.setDefaultWidget(select_all_widget)
        menu.addAction(select_all_action)
        menu.addSeparator()
        select_all_widget.adjustSize()
        max_width = max(max_width, select_all_widget.sizeHint().width())

        # Add value checkboxes using QWidgetAction
        for value in values:
            cb = QCheckBox(str(value), menu)
            cb.setChecked(value in self.active_filter_values.get(key, set(values)))
            action = QWidgetAction(menu)
            action.setDefaultWidget(cb)
            menu.addAction(action)
            actions[value] = cb
            cb.adjustSize()
            max_width = max(max_width, cb.sizeHint().width())

        def update_select_all_state():
            all_checked = all(cb.isChecked() for cb in actions.values())
            select_all_widget.blockSignals(True)
            select_all_widget.setChecked(all_checked)
            select_all_widget.blockSignals(False)

        def update_filters():
            selected_values = [v for v, cb in actions.items() if cb.isChecked()]
            self.active_filter_values[key] = set(selected_values)
            self.proxyModel.set_filter(key, selected_values)
            update_select_all_state()

        for value, cb in actions.items():
            cb.toggled.connect(update_filters)

        def on_select_all_toggled(checked):
            for cb in actions.values():
                cb.blockSignals(True)
                cb.setChecked(checked)
                cb.blockSignals(False)
            update_filters()

        select_all_widget.toggled.connect(on_select_all_toggled)
        update_select_all_state()

        # Set menu width to fit content
        menu.setFixedWidth(max_width + 5)

        # When menu is about to hide, if nothing is checked, check all and update filter
        def on_menu_about_to_hide():
            if not any(cb.isChecked() for cb in actions.values()):
                for cb in actions.values():
                    cb.blockSignals(True)
                    cb.setChecked(True)
                    cb.blockSignals(False)
                update_filters()
            self._current_menu = None

        menu.aboutToHide.connect(on_menu_about_to_hide)
        return menu
        
    def open_filter_popup(self, key):
        menu = self.filter_menus[key]
        button = self.filter_buttons[key]
        self._current_menu = menu
        menu.setMinimumWidth(button.width())
        menu.exec_(button.mapToGlobal(button.rect().bottomLeft()))
        
    def model(self):
        return self.proxyModel

# Simple test window with default data
def main():
    app = QApplication(sys.argv)
    stringPropertiesData_list = [
        {"name": "Apple", "type": "Fruit", "color": "red"},
        {"name": "Banana", "type": "Fruit", "color": "yellow"},
        {"name": "Carrot", "type": "Vegetable", "color": "orange"},
        {"name": "Date", "type": "Fruit", "color": "orange"},
        {"name": "Eggplant", "type": "Vegetable", "color": "yellow"}
    ]
    model = QStringPropertyListModel()
    model.setDataFromDict(stringPropertiesData_list, string_key="name", properties_keys=["type", "color"])
    window = QWidget()
    layout = QVBoxLayout(window)
    filterable = QFilterableComboBoxView(label_text="Select:", parent=window)
    filterable.setModel(model)
    layout.addWidget(filterable)
    window.setWindowTitle("Filterable ComboBox Test")
    window.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
"""
Created on 29.11.2020

@author: Martin
"""

from pandas import DataFrame
import koalafolio.gui.helper.QLogger as logger

localLogger = logger.globalLogger

# staic class for static data
class ChaindataStatic:
    BLOCKDAEMON_BASE_URL = "https://svc.blockdaemon.com/"
    SUPPORTED_CHAINS = [
        "cardano",
        "ethereum",
        "solana",
        "polkadot",
        "polygon",
        "near",
        "avalanche"
    ]
    apiBaseURLs = {"cardano": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/cardano/mainnet/delegator/history/",
                   "ethereum": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/ethereum/mainnet/validator/history/",
                   "solana": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/solana/mainnet/delegator/history/",
                   "polkadot": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/polkadot/mainnet/nominator/history/",
                   "polygon": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/polygon/mainnet/delegator/history/",
                   "near": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/near/mainnet/delegator/history/",
                   "avalanche": BLOCKDAEMON_BASE_URL + "reporting/staking/v1/avalanche/mainnet/delegator/history/"}

    # time format factor, s = 1, ms = 1000, ...
    apiTimeFactors = {"cardano": 1, "ethereum": 1000, "solana": 1000, "polkadot": 1000, "polygon": 1, "near": 1,
                      "avalanche": 1}

    # timeUnit
    apiTimeUnits = {"cardano": "epoch", "ethereum": "week", "solana": "epoch", "polkadot": "era", "polygon": "epoch",
                    "near": "weekly", "avalanche": ""}

    @staticmethod
    def getBlockdaemonRewardsForAddress(apiname: str, apikey: str, address: str, start: int,
                                        end: int) -> DataFrame:
        startTimestamp = int(start) * ChaindataStatic.apiTimeFactors[apiname]
        endTimestamp = int(end) * ChaindataStatic.apiTimeFactors[apiname]
        if ChaindataStatic.apiTimeUnits[apiname]:
            payload = {
                "fromTime": startTimestamp,
                "toTime": endTimestamp,
                "timeUnit": ChaindataStatic.apiTimeUnits[apiname]
            }
        else:
            payload = {
                "fromTime": startTimestamp,
                "toTime": endTimestamp
            }

        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "X-API-Key": apikey
        }

        from requests import post as requests_post
        response = requests_post(ChaindataStatic.apiBaseURLs[apiname] + str(address), json=payload, headers=headers)
        try:
            content = response.json()
        except Exception as ex:
            localLogger.error("no data returned from Blockdaemon (" + str(apiname) + "): " + str(ex))
            return DataFrame()
        if "rewards" in content:
            rewards = content['rewards']
        else:
            rewards = content

        return DataFrame.from_dict(rewards)





# -*- coding: utf-8 -*-
"""
Created on Wen Jan 02 14:21:00 2019

@author: Martin
"""

import os.path
from PyQt5.QtCore import QAbstractListModel, QModelIndex, QObject, QTimer, QVariant, Qt, pyqtSignal
from PyQt5.QtGui import QBrush, QColor, QMouseEvent
from PyQt5.QtWidgets import QAbstractItemView, QListView
import koalafolio.gui.widgets.ScrollableTable as sTable

class QLogModel(QAbstractListModel):
    def __init__(self, *args, **kwargs):
        super(QLogModel, self).__init__(*args, **kwargs)
        self.stringList = []
        self.messageType = []
        self.color = []

        self.errorColorBrush = QBrush(QColor(255, 0, 0, 255))
        self.warningColorBrush = QBrush(QColor(255, 165, 0, 255))

    def addString(self, message, messageType):
        row = len(self.stringList)
        self.beginInsertRows(QModelIndex(), row, row+1)
        self.stringList.append(message)
        self.messageType.append(messageType)

        # infoColor = QColor(0, 0, 0, 255)
        if messageType == 'e':
            self.color.append(self.errorColorBrush)
        elif messageType == 'w':
            self.color.append(self.warningColorBrush)
        else:
            self.color.append(QVariant())
        self.endInsertRows()

    def rowCount(self, parent : QModelIndex) -> int:
        return len(self.stringList)

    def data(self, index: QModelIndex, role: int):
        if role == Qt.DisplayRole:
            return self.stringList[index.row()]
        if role == Qt.ForegroundRole:
            return self.color[index.row()]
        return QVariant()

    def nextIndex(self, currentIndex):
        return self.index(currentIndex.row()+1, 0, currentIndex.parent())



class QLogView(QListView):

    def __init__(self, parent=None, logfile=None, *args, **kwargs):
        super(QLogView, self).__init__(parent=parent, *args, **kwargs)

        self.logfile = logfile
        self.setVerticalScrollBar(sTable.MinWheelScrollingScrollbar(orientation=Qt.Vertical, parent=self))

        self.autoScrollTimer = TimedAutoScroll()
        self.autoScrollTimer.performScrollStep.connect(self.scrollStep)

        self.currentLastRow = 0
        self.scrollBarRowOffset = 3

    def setModel(self, model, *args, **kwargs):
        self.currentLastRow = model.rowCount(parent=QModelIndex()) - 1
        self.verticalScrollBar().setValue(1)
        self.scrollBarRowOffset = self.currentLastRow - self.verticalScrollBar().value()
        super(QLogView, self).setModel(model, *args, **kwargs)

    def rowsInserted(self, parent, start, end):
        super(QLogView, self).rowsInserted(parent, start, end)
        self.autoScrollTimer.addScrollSteps(end-start)
        #self.scrollToBottom()

    def scrollStep(self):
        self.currentLastRow += 1
        if self.isAtCurrentLastRow():
            index = self.model().index(self.currentLastRow, 0, QModelIndex())
            self.scrollTo(index, QAbstractItemView.PositionAtBottom)

    def isAtCurrentLastRow(self):
        #print("currentLastRow: " + str(self.currentLastRow) + ", scrollBarValue: " + str(self.verticalScrollBar().value() + self.scrollBarRowOffset))
        return -1 <= (self.verticalScrollBar().value() + self.scrollBarRowOffset) - self.currentLastRow <= 1

    def mouseDoubleClickEvent(self, a0: QMouseEvent) -> None:
        try:
            os.system(self.logfile)
        except Exception:
            return


class TimedAutoScroll(QObject):
    performScrollStep = pyqtSignal()

    def __init__(self):
        super(TimedAutoScroll, self).__init__()
        self.scrollStepCount = 0
        self.isScrollQueueEmpty = True

        self.scrollTimer = QTimer()
        self.scrollTimer.timeout.connect(self.triggerStep)
        self.scrollTimer.start(300)

    def addScrollSteps(self, count):
        self.scrollStepCount += count
        if self.isScrollQueueEmpty:
            self.triggerStep()

    def triggerStep(self):
        if self.scrollStepCount > 0:
            self.isScrollQueueEmpty = False
            self.scrollStepCount -= 1
            self.performScrollStep.emit()
        else:
            self.isScrollQueueEmpty = True



# -*- coding: utf-8 -*-
"""
Lazy Page Loader for optimized startup performance

Created for koalafolio optimization
"""

from PyQt5.QtCore import QObject, QTimer, Qt, pyqtSignal
from PyQt5.QtWidgets import QLabel, QVBoxLayout, QWidget


class LazyPageLoader(QObject):
    """Handles lazy loading of GUI pages to improve startup performance"""

    pageLoaded = pyqtSignal(int, QWidget)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.loaded_pages = {}
        self.loading_pages = set()

    def loadPage(self, page_index, page_class_name):
        """Load a page lazily when first accessed"""
        if page_index in self.loaded_pages:
            return self.loaded_pages[page_index]

        if page_index in self.loading_pages:
            return None  # Already loading

        self.loading_pages.add(page_index)

        # Use QTimer to load page in next event loop iteration
        QTimer.singleShot(0, lambda: self._loadPageAsync(page_index, page_class_name))

        return None

    def _loadPageAsync(self, page_index, page_class_name):
        """Actually load the page asynchronously"""
        try:
            # Import and create the page
            if page_class_name == 'PortfolioPage':
                from koalafolio.gui.pages.PortfolioPage import PortfolioPage
                page = PortfolioPage(parent=self.parent, controller=self.parent)
            elif page_class_name == 'TradesPage':
                from koalafolio.gui.pages.TradesPage import TradesPage
                page = TradesPage(parent=self.parent, controller=self.parent)
            elif page_class_name == 'ImportPage':
                from koalafolio.gui.pages.ImportPage import ImportPage
                page = ImportPage(parent=self.parent, controller=self.parent)
            elif page_class_name == 'ExportPage':
                import koalafolio.gui.pages.ExportPage as exportpage
                page = exportpage.ExportPage(parent=self.parent, controller=self.parent)
            elif page_class_name == 'SettingsPage':
                import koalafolio.gui.pages.Qpages as pages
                page = pages.SettingsPage(parent=self.parent, controller=self.parent)
            else:
                raise ValueError(f"Unknown page class: {page_class_name}")

            self.loaded_pages[page_index] = page
            self.loading_pages.discard(page_index)
            self.pageLoaded.emit(page_index, page)

        except Exception as ex:
            print(f"Error loading page {page_class_name} at index {page_index}: {ex}")
            import traceback
            traceback.print_exc()
            self.loading_pages.discard(page_index)

    def getPage(self, page_index):
        """Get a loaded page or None if not loaded yet"""
        return self.loaded_pages.get(page_index)

    def isPageLoaded(self, page_index):
        """Check if a page is loaded"""
        return page_index in self.loaded_pages

    def isPageLoading(self, page_index):
        """Check if a page is currently loading"""
        return page_index in self.loading_pages


class PlaceholderPage(QWidget):
    """Placeholder widget shown while pages are loading"""

    def __init__(self, page_name="Page", parent=None):
        super().__init__(parent)
        self.page_name = page_name
        self.setupUI()

    def setupUI(self):
        layout = QVBoxLayout(self)

        # Add loading indicator with larger font
        loading_label = QLabel(f"Loading {self.page_name}...")
        loading_label.setAlignment(Qt.AlignCenter)
        font = loading_label.font()
        font.setPointSize(14)
        font.setBold(True)
        loading_label.setFont(font)

        # Add progress indicator (simple animated dots)
        self.progress_label = QLabel("●○○")
        self.progress_label.setAlignment(Qt.AlignCenter)
        progress_font = self.progress_label.font()
        progress_font.setPointSize(16)
        self.progress_label.setFont(progress_font)

        layout.addStretch()
        layout.addWidget(loading_label)
        layout.addWidget(self.progress_label)
        layout.addStretch()

        # Animate the progress indicator
        self.animation_timer = QTimer()
        self.animation_timer.timeout.connect(self.updateProgress)
        self.animation_timer.start(500)  # Update every 500ms
        self.animation_state = 0

    def updateProgress(self):
        """Update the progress animation"""
        states = ["●○○", "○●○", "○○●", "○●○"]
        self.progress_label.setText(states[self.animation_state])
        self.animation_state = (self.animation_state + 1) % len(states)

    def refresh(self):
        """Placeholder refresh method"""
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test script for exchange API functions in koalafolio/web/exchanges.py
Tests all getTradeHistory functions with API keys from a text file
"""

import os
import time
import datetime
import pandas as pd
import ast
from typing import Dict
import unittest
from pathlib import Path
from koalafolio.web.exchanges import ExchangesStatic

# Add parent directory to path to import koalafolio modules
project_root = Path(__file__).parents[1]
test_root = Path(__file__).parents[2] / "koalafolio_testdata"
testdata_path = test_root / Path(__file__).stem


class TestExchanges(unittest.TestCase):
    """Test class for exchange API functions"""

    def setUp(self):
        """Set up test environment"""
        self.test_start_time = datetime.datetime.now()
        self.test_data_dir = testdata_path
        # Check if test data directory exists
        if not os.path.exists(self.test_data_dir):
            raise unittest.SkipTest(f"Test data directory not found: {self.test_data_dir}")
        self.test_data_out_dir = testdata_path / f"out_{self.test_start_time.strftime('%Y%m%d_%H%M%S')}"
        # create out dir if it does not exist
        if not os.path.exists(self.test_data_out_dir):
            os.makedirs(self.test_data_out_dir)
        self.output_file = self.test_data_out_dir / \
                           f"test_exchanges_results_{self.test_start_time.strftime('%Y%m%d_%H%M%S')}.json"

        # Calculate timestamps for 5 year span
        now = self.test_start_time
        some_years_ago = now - datetime.timedelta(days=5*365)
        
        self.end_ts = int(time.mktime(now.timetuple()))
        self.start_ts = int(time.mktime(some_years_ago.timetuple()))

        self.load_api_keys(self.test_data_dir /"input.csv")

        # self.settings = settings.mySettings.setPath(test_root)

    def load_api_keys(self, file_path: str) -> Dict[str, Dict[str, str]]:
        """Load API keys from text file
        
        Args:
            file_path: Path to text file with API keys
            
        """
        # Dictionary to store API keys
        self.api_keys = {}
        try:
            with open(file_path, 'r') as f:
                content = f.read()
                self.api_keys = ast.literal_eval(content)
            
            print(f"TEST test_exchanges: Loaded {len(self.api_keys)} API keys from {file_path}")
        except Exception as e:
            print(f"TEST test_exchanges: Error loading API keys: {e}")
    
    def save_trades_to_csv(self, exchange_name: str, trades_df: pd.DataFrame):
        """Save trades to CSV file
        
        Args:
            exchange_name: Name of the exchange
            trades_df: DataFrame with trades
        """
        if not trades_df.empty:
            filename = f"{exchange_name}_ccxt_trades.txt"
            
            trades_df.to_csv(self.test_data_out_dir / filename, index=False)
            print(f"TEST test_exchanges: Saved {len(trades_df)} trades to {self.test_data_out_dir / filename}")
            
    def runTest(self):
        """Run all tests"""
        for apiname in self.api_keys:
            api_data = self.api_keys[apiname]
            key = api_data['apikey']
            secret = api_data['secret']
            print(f"TEST test_exchanges: Testing {apiname} API...")
            try:
                # log time needed for getTradeHistoryCall
                start_time = time.time()
                result = ExchangesStatic.getTradeHistoryCcxt(apiname, key, secret, self.start_ts, self.end_ts)
                end_time = time.time()
                print(f"TEST test_exchanges: {apiname} API test successful. Got {len(result)} trades.")
                self.save_trades_to_csv(apiname, result)
                print(f"TEST test_exchanges: {apiname} API test took {end_time - start_time:.2f} seconds")
                self.assertIsInstance(result, pd.DataFrame)
            except Exception as e:
                print(f"TEST test_exchanges: {apiname} API test failed: {e}")
                # mark test as failed
                self.assertTrue(False)




if __name__ == "__main__":
    # run test with test data
    unittest.main()



# -*- coding: utf-8 -*-
"""
Test script for converter functions in koalafolio/Import/Converter.py
Tests all modelCallback functions with example trade data
"""

import os
import pandas as pd
import json
import datetime
import unittest
from pathlib import Path
import koalafolio.Import.Models as models
import koalafolio.PcpCore.core as core
import koalafolio.PcpCore.settings as settings

# Add parent directory to path to import koalafolio modules
project_root = Path(__file__).parents[1]
test_root = Path(__file__).parents[2] / "koalafolio_testdata"
testdata_path = test_root / Path(__file__).stem


class ConverterTestResult :
    def __init__(self, model_name, file_name, success, trades_count=0, fees_count=0, skipped_rows=0, trade_list=None, fee_list=None, error=None):
        self.model_name = model_name
        self.file_name = file_name
        self.success = success
        self.trades_count = trades_count
        self.fees_count = fees_count
        self.skipped_rows = skipped_rows
        self.error = error
        self.timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.trade_list = trade_list
        self.fee_list = fee_list
    
    def to_dict(self):
        return {
            "model_name": self.model_name,
            "file_name": self.file_name,
            "success": self.success,
            "trades_count": self.trades_count,
            "fees_count": self.fees_count,
            "skipped_rows": self.skipped_rows,
            "error": str(self.error) if self.error else None,
            "timestamp": self.timestamp
            # skip trade_list and fee_list in dict creation
        }

class TestConverter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Set up test fixtures before running tests."""
        cls.start_time = datetime.datetime.now()
        cls.test_data_dir = testdata_path
        # Check if test data directory exists
        if not os.path.exists(cls.test_data_dir):
            raise unittest.SkipTest(f"Test data directory not found: {cls.test_data_dir}")
        cls.test_data_out_dir = testdata_path / f"out_{cls.start_time.strftime('%Y%m%d_%H%M%S')}"
        # create out dir if it does not exist
        if not os.path.exists(cls.test_data_out_dir):
            os.makedirs(cls.test_data_out_dir)
        cls.test_spec_file = "testspec.txt"
        cls.output_file = cls.test_data_out_dir / \
                          f"test_converter_results_{cls.start_time.strftime('%Y%m%d_%H%M%S')}.json"
        cls.results = []
        cls.settings = settings.mySettings.setPath(test_root)

    def get_test_files(self):
        """Get list of files to test."""
        if os.path.exists(os.path.join(self.test_data_dir, self.test_spec_file)):
            with open(os.path.join(self.test_data_dir, self.test_spec_file), 'r') as f:
                # return all files from test_spec_file if not commented out
                return [line.strip() for line in f if line.strip() and not line.startswith('#')]
        # return all files in the test_data_dir
        return [f for f in os.listdir(self.test_data_dir) if os.path.isfile(os.path.join(self.test_data_dir, f)) and not f.startswith('.')]

    def runTest(self):
        """Test all converters with test data files."""
        for file_name in self.get_test_files():
            file_path = os.path.join(self.test_data_dir, file_name)
            if not os.path.exists(file_path):
                print(f"TEST test_converter: File not found: {file_path}")
                continue
            
            try:
                df = pd.read_csv(file_path)
                matchcnt = 0
                for model in models.IMPORT_MODEL_LIST:
                    try:
                        if model.isMatch(df.columns.tolist()):
                            try:
                                trade_list, fee_list, skipped_rows = model.convertDataFrame(df)

                                if len(trade_list) > 0 or len(fee_list) > 0:
                                    matchcnt += 1
                                    self.results.append(ConverterTestResult (
                                        model_name=model.modelName or model.modelCallback.__name__,
                                        file_name=file_name,
                                        success=True,
                                        trades_count=len(trade_list),
                                        fees_count=len(fee_list),
                                        skipped_rows=skipped_rows,
                                        trade_list=trade_list,
                                        fee_list=fee_list
                                    ))
                                    print(
                                        f"TEST test_converter: Success: {file_name} with {model.modelCallback.__name__}")
                                    print(
                                        f"TEST test_converter:   Trades: {len(trade_list)}, Fees: {len(fee_list)}, Skipped: {skipped_rows}")
                                else:
                                    self.results.append(ConverterTestResult (
                                        model_name=model.modelName or model.modelCallback.__name__,
                                        file_name=file_name,
                                        success=True,
                                        trades_count=0,
                                        fees_count=0,
                                        skipped_rows=skipped_rows,
                                        error="No trades or fees found"
                                    ))
                                    print(
                                        f"TEST test_converter: Info: {file_name} matched with {model.modelCallback.__name__} but no trades or fees found")
                                    print(
                                        f"TEST test_converter:   Trades: {len(trade_list)}, Fees: {len(fee_list)}, Skipped: {skipped_rows}")

                                # Add assertions to verify the conversion results
                                self.assertIsInstance(trade_list, core.TradeList)
                                self.assertIsInstance(fee_list, core.TradeList)
                                self.assertIsInstance(skipped_rows, int)
                                
                            except Exception as e:
                                self.results.append(ConverterTestResult (
                                    model_name=model.modelName or model.modelCallback.__name__,
                                    file_name=file_name,
                                    success=False,
                                    error=e
                                ))
                                print(f"TEST test_converter: Conversion error: {file_name} with {model.modelCallback.__name__}: {str(e)}")
                    except Exception as e:
                        self.results.append(ConverterTestResult (
                            model_name=model.modelName or model.modelCallback.__name__,
                            file_name=file_name,
                            success=False,
                            error=e
                        ))
                        print(f"TEST test_converter: Matching error: {file_name} with {model.modelCallback.__name__}: {str(e)}")
                if matchcnt == 0:
                    self.results.append(ConverterTestResult (
                        model_name="file_reading",
                        file_name=file_name,
                        success=False,
                        error="No matching model found"
                    ))
                    print(f"TEST test_converter: Matching error: {file_name} was not matched by any model")
                if matchcnt > 1:
                    fileresults = [result for result in self.results if result.file_name == file_name]
                    # check if any fileresult has different amount of trades, fees or skippedrows
                    if len(set([result.trades_count for result in fileresults])) > 1 or \
                            len(set([result.fees_count for result in fileresults])) > 1 or \
                            len(set([result.skipped_rows for result in fileresults])) > 1:
                        self.results.append(ConverterTestResult (
                            model_name="file_reading",
                            file_name=file_name,
                            success=False,
                            error="Multiple matching models found with different results"
                        ))
                        print(f"TEST test_converter: Matching error: {file_name} has multiple matching models with different results")
                    else:
                        self.results.append(ConverterTestResult (
                            model_name="file_reading",
                            file_name=file_name,
                            success=True,
                            error="Multiple matching models found with equal output"
                        ))
                        print(f"TEST test_converter: Warning: {file_name} has multiple matching with equal results")
            except Exception as e:
                self.results.append(ConverterTestResult (
                    model_name="file_reading",
                    file_name=file_name,
                    success=False,
                    error=e
                ))
                print(f"TEST test_converter: File reading error: {file_path}: {str(e)}")
        
        # Write results to file
        self.write_results()
        
        # Verify that we have some successful tests
        success_count = sum(1 for r in self.results if r.success)
        total_count = len(self.results)
        print(f"TEST test_converter: Summary: {success_count}/{total_count} tests passed")
        self.assertGreater(success_count, 0, "No successful conversions")

    def write_results(self):
        """Write test results to a JSON file."""
        results_dict = [result.to_dict() for result in self.results]
        
        grouped_results = {}
        for result in results_dict:
            file_name = result["file_name"]
            if file_name not in grouped_results:
                grouped_results[file_name] = []
            grouped_results[file_name].append(result)
        
        with open(self.output_file, 'w') as f:
            json.dump(grouped_results, f, indent=2)

        for result in self.results:
            if result.success:
                trade_list, fee_list = result.trade_list, result.fee_list
                if trade_list:
                    tradeFile = self.test_data_out_dir / f"{result.file_name.split('.')[0]}_{result.model_name}_trades.txt"
                    trade_list.toCsv(tradeFile)
                if fee_list:
                    feeFile = self.test_data_out_dir / f"{result.file_name.split('.')[0]}_{result.model_name}_fees.txt"
                    fee_list.toCsv(feeFile)

        
        print(f"TEST test_converter: Test results written to {self.output_file}")

if __name__ == "__main__":
    # run test with unittest
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from PyQt5.QtWidgets import QComboBox, QCompleter
from PyQt5.QtCore import QStringListModel, Qt, QEvent, QSortFilterProxyModel, QTimer


class StyledCompleter(QCompleter):
    # Class-level storage for all instances and shared style
    _instances = set()
    _shared_stylesheet = ""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.stylesheet = ""
        # Register instance
        StyledCompleter._instances.add(self)
        # Apply current shared style if exists
        if StyledCompleter._shared_stylesheet:
            self.setStylesheet(StyledCompleter._shared_stylesheet)
        self.stylePopup()

    def __del__(self):
        # Clean up instance registration
        StyledCompleter._instances.discard(self)

    @classmethod
    def updateAllStyles(cls, style):
        """Class method to update all instances with new style"""
        cls._shared_stylesheet = style
        for instance in cls._instances:
            instance.setStylesheet(style)

    def setStylesheet(self, style):
        self.stylesheet = style
        self.stylePopup()

    def stylePopup(self):
        # delegate = QStyledItemDelegate(self)
        popup = self.popup()
        # popup.setItemDelegate(delegate)
        popup.setObjectName("StyledCompleterPopup")  # Allows styling via QSS
        # popup.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
        # popup.setAutoFillBackground(False)
        popup.setStyleSheet(self.stylesheet)

    def setModel(self, model):
        super().setModel(model)
        self.stylePopup()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Escape:
            try:
                self.parent().lineEdit().clear()
            except AttributeError:
                pass
        return super().eventFilter(obj, event)


class PrefixFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._filter_text = ""

    def setFilterText(self, text):
        self._filter_text = text
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        if not self._filter_text:
            return True
        index = self.sourceModel().index(source_row, 0, source_parent)
        data = self.sourceModel().data(index, Qt.DisplayRole)
        return data.lower().startswith(self._filter_text.lower())


class QCompleterComboBoxView(QComboBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setEditable(True)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)  # Prevent adding new items

        # Set up proxy filter model
        self.proxyModel = PrefixFilterProxyModel(self)

        # Set up completer
        completer = StyledCompleter(parent=self)
        completer.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)  # Case-insensitive matching
        completer.setModel(self.proxyModel)
        self.setCompleter(completer)

        # Connect text input changes and enter key press
        self.lineEdit().textEdited.connect(self._handle_text_edited)
        self.lineEdit().returnPressed.connect(self.select_first_completion)

        self.activated.connect(self._on_activated)
        self.currentIndexChanged.connect(self._on_current_index_changed)

    # def currentText(self):
    #     # call data with current index
    #     return self.model().data(self.model().index(self.currentIndex(), 0), Qt.DisplayRole)

    def _on_activated(self, index):
        # Get the display text from the model and set it to the line edit
        text = self.model().data(self.model().index(index, 0), Qt.DisplayRole)
        self.lineEdit().setText(text)

    def _on_current_index_changed(self, index):
        # Get the display text from the model and set it to the line edit
        text = self.model().data(self.model().index(index, 0), Qt.DisplayRole)
        self.lineEdit().setText(text)

    def addItem(self, text, userData = ...):
        # raise notImplemented
        raise NotImplementedError("Use setModel() to set a model with items.")

    def addItems(self, items):
        # raise notImplemented
        raise NotImplementedError("Use setModel() to set a model with items.")

    def setModel(self, model):
        super().setModel(model)
        self.proxyModel.setSourceModel(model)

    def _handle_text_edited(self, text):
        # Only show completer popup if user is actually typing
        if self.lineEdit().hasFocus():
            self.update_filter(text)
            # process completer after the event loop has processed the text change
            # self._show_completer()
            QTimer.singleShot(0, self._show_completer)

    def _show_completer(self):
        self.completer().complete()

    def update_filter(self, text: str):
        self.proxyModel.setFilterText(text)

    def getFirstCompletion(self):
        if self.proxyModel.rowCount() > 0:
            index = self.proxyModel.index(0, 0)
            return self.proxyModel.data(index, Qt.DisplayRole)
        return ""

    def select_first_completion(self):
        completer = self.completer()
        if not completer:
            return
        completerModel = completer.completionModel()
        if not completerModel or completerModel.rowCount() == 0:
            self.clearEditText()
            return
        index = completer.popup().currentIndex()
        if not index.isValid():
            index = completerModel.index(0, 0)
        if index.isValid():
            # Replace with suggestion
            data = completerModel.data(index, Qt.DisplayRole)
            QTimer.singleShot(0, lambda: self.setCurrentText(data))

    def showPopup(self):
        # update current index based on current text
        index = self.findText(self.lineEdit().text())
        if index < 0:
            index = self.findText(self.getFirstCompletion())
        if index >= 0:
            self.setCurrentIndex(index)
        super().showPopup()

    def focusOutEvent(self, event):
        super().focusOutEvent(event)
        # try to set completer text if current text is not in model
        if self.findText(self.currentText()) == -1:
            self.select_first_completion()

# create a test window with the QCompleterComboBoxView and a dummy QStringListModel
# for testing and run it if this file is run directly
if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout
    from PyQt5.QtCore import QStringListModel

    app = QApplication([])

    window = QWidget()
    layout = QVBoxLayout(window)

    model = QStringListModel(["Apple", "Banana", "Cherry", "Date", "Fig", "Grape"])
    combo = QCompleterComboBoxView()
    combo.setModel(model)

    layout.addWidget(combo)
    window.setLayout(layout)
    window.show()

    app.exec_()




# -*- coding: utf-8 -*-
"""
Created on Sun Sep 16 09:57:14 2018

@author: Martin
"""

# Core imports - needed immediately
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import (QApplication, QWidget, QVBoxLayout,
                             QHBoxLayout, QSplashScreen)
from PyQt5.QtCore import Qt, QTimer, QThread, pyqtSignal, QSize, QRect
from sys import argv, exit

# Essential application imports
import koalafolio.PcpCore.logger as coreLogger
import koalafolio.gui.helper.QLogger as logger
import koalafolio.PcpCore.arguments as arguments
import koalafolio.PcpCore.settings as coreSettings
import koalafolio.gui.helper.QSettings as settings
import koalafolio.gui.helper.QStyle as style
import koalafolio.exp.QTranslate as translator

from koalafolio.gui.widgets.QCompleterComboBox import StyledCompleter
# Defer heavy imports - will be imported when needed
# import koalafolio.gui.QLoggerWidget as loggerwidget
# import koalafolio.gui.Qpages as pages
# import koalafolio.gui.QExportPage as exportpage
# import koalafolio.gui.Qcontrols as controls
# import koalafolio.gui.QPortfolioTable as ptable
# import koalafolio.gui.QTradeTable as ttable
# import koalafolio.gui.QThreads as threads
# import koalafolio.gui.SettingsPage as settingsPage
# import koalafolio.Import.apiImport as apiImport
# import requests
# from koalafolio.gui.TradesPage import TradesPage
# from koalafolio.gui.ImportPage import ImportPage
# from koalafolio.gui.PortfolioPage import PortfolioPage


# %% constants


# %% Custom splash screen with better text positioning
class CustomSplashScreen(QSplashScreen):
    """Custom splash screen with improved text positioning"""

    def __init__(self, pixmap):
        super().__init__(pixmap)
        self.message = ""
        self.text_color = Qt.white
        self.text_alignment = Qt.AlignBottom | Qt.AlignCenter

        # Set larger font (30% bigger)
        font = self.font()
        current_size = font.pointSize()
        if current_size <= 0:
            current_size = 12
        new_size = int(current_size * 1.3)
        font.setPointSize(new_size)
        font.setBold(True)
        self.setFont(font)

    def showMessage(self, message, alignment=None, color=None):
        """Override to store message and trigger repaint"""
        self.message = message
        if alignment is not None:
            self.text_alignment = alignment
        if color is not None:
            self.text_color = color
        self.repaint()

    def drawContents(self, painter):
        """Custom drawing to position text higher"""
        if not self.message:
            return

        painter.setPen(self.text_color)
        painter.setFont(self.font())

        # Calculate text position - move up by one text height
        font_metrics = painter.fontMetrics()
        text_height = font_metrics.height()

        # Get the rectangle for drawing
        rect = self.rect()

        # Move the bottom up by one text height plus some padding
        adjusted_rect = QRect(rect.x(), rect.y(),
                                   rect.width(), rect.height() - text_height - 10)

        painter.drawText(adjusted_rect, self.text_alignment, self.message)


# %% Background version check thread
class VersionCheckThread(QThread):
    versionChecked = pyqtSignal(str)

    def __init__(self, package_name='koalafolio'):
        super().__init__()
        self.package_name = package_name

    def run(self):
        try:
            import requests
            response = requests.get(f'https://pypi.org/pypi/{self.package_name}/json', timeout=5)
            latest_version = str(response.json()['info']['version'])
            self.versionChecked.emit(latest_version)
        except Exception as ex:
            logger.globalLogger.warning(f'Could not check latest version: {ex}')
            self.versionChecked.emit('')


# %% Lazy import koalafolio.gui.helper functions
def lazy_import_heavy_modules():
    """Import heavy modules when actually needed"""
    global loggerwidget, pages, exportpage, controls, ptable, ttable, threads
    global settingsPage, apiImport
    global TradesPage, ImportPage, PortfolioPage

    import koalafolio.gui.widgets.QLoggerWidget as loggerwidget
    import koalafolio.gui.pages.Qpages as pages
    import koalafolio.gui.pages.ExportPage as exportpage
    import koalafolio.gui.widgets.Qcontrols as controls
    import koalafolio.gui.widgets.QPortfolioTable as ptable
    import koalafolio.gui.widgets.QTradeTable as ttable
    import koalafolio.gui.QThreads as threads
    import koalafolio.gui.pages.SettingsPage as settingsPage
    import koalafolio.Import.apiImport as apiImport
    from koalafolio.gui.pages.TradesPage import TradesPage
    from koalafolio.gui.pages.ImportPage import ImportPage
    from koalafolio.gui.pages.PortfolioPage import PortfolioPage


# %% variables
import sys
import os

if getattr(sys, 'frozen', False):  # running as pyinstaller executable
    application_path = os.path.dirname(sys.executable)
    internal_dir = os.path.join(application_path, "_internal")  # Check for _internal
    running_mode = 'Frozen/executable'
else:
    try:
        app_full_path = os.path.realpath(__file__)
        application_path = os.path.dirname(app_full_path)
        running_mode = "Non-interactive"
    except NameError:
        application_path = os.getcwd()
        running_mode = 'Interactive'

def getResourcePath(relative_path) -> str:
    """Get absolute path to resource, works for dev and for PyInstaller"""
    base_path = getattr(sys, '_MEIPASS', application_path)
    if running_mode == 'Frozen/executable':
        base_path = internal_dir
        # Check where the files are stored
        internalPath = os.path.join(internal_dir, relative_path)
        if os.path.exists(internalPath):
            return internalPath
        else:
            return os.path.join(application_path, relative_path)  # Try the exe folder
    return os.path.join(base_path, relative_path)

# %% classes
class PortfolioApp(QWidget):
    PORTFOLIOPAGEINDEX = 0
    TRADESPAGEINDEX = 1
    IMPORTPAGEINDEX = 2
    EXPORTPAGEINDEX = 3
    SETTINGSPAGEINDEX = 4
    startRefresh = pyqtSignal()
    endRefresh = pyqtSignal()
    styleSheetChanged = pyqtSignal(str)

    def __init__(self, parent=None, userDataPath = "", username="", splash=None):
        super(PortfolioApp, self).__init__(parent=parent)

        self.userDataPath = userDataPath
        self.username = username
        self.splash = splash
        self.latestVersion = ""
        self.heavy_modules_loaded = False
        self.data_loaded = False

        self.setObjectName("MainWindow")

        # Update splash
        if self.splash:
            self.splash.showMessage("Initializing environment...", Qt.AlignBottom | Qt.AlignCenter)
            QApplication.processEvents()

        self.initEnv()

        if self.splash:
            self.splash.showMessage("Setting up window...", Qt.AlignBottom | Qt.AlignCenter)
            QApplication.processEvents()

        self.initWindow()

        # Start version check in background (non-blocking)
        self.versionCheckThread = VersionCheckThread()
        self.versionCheckThread.versionChecked.connect(self.onVersionChecked)
        self.versionCheckThread.start()

        if self.splash:
            self.splash.showMessage("Loading modules...", Qt.AlignBottom | Qt.AlignCenter)
            QApplication.processEvents()

        # Load heavy modules
        lazy_import_heavy_modules()
        self.heavy_modules_loaded = True

        if self.splash:
            self.splash.showMessage("Initializing data models...", Qt.AlignBottom | Qt.AlignCenter)
            QApplication.processEvents()

        # init data
        self.initData()

        if self.splash:
            self.splash.showMessage("Setting up interface...", Qt.AlignBottom | Qt.AlignCenter)
            QApplication.processEvents()

        # init style
        self.initStyle()

        # init gui
        self.layoutUI()

        if self.splash:
            self.splash.showMessage("Starting application...", Qt.AlignBottom | Qt.AlignCenter)
            QApplication.processEvents()

        self.show()
        self.showFrame(self.PORTFOLIOPAGEINDEX)

        # Start background initialization
        self.startBackgroundInitialization()

        self.logger.info('application initialized')

    def onVersionChecked(self, latest_version):
        """Handle version check result"""
        self.latestVersion = latest_version
        if latest_version and self.settings['general']['version'] != latest_version:
            self.logger.info('new version ' + latest_version + ' available. Install with "pip install koalafolio --upgrade"')

    def startBackgroundInitialization(self):
        """Start background threads and data loading"""
        # Use QTimer to defer heavy operations until after UI is shown
        QTimer.singleShot(100, self.loadDataInBackground)

    def loadDataInBackground(self):
        """Load data in background after UI is ready"""
        try:
            # start the background threads
            self.startThreads()

            # Load data asynchronously
            QTimer.singleShot(200, lambda: self.coinList.restoreCoins())
            QTimer.singleShot(300, lambda: self.tradeList.restoreTrades())

            self.data_loaded = True
            self.logger.info('background data loading completed')
        except Exception as ex:
            self.logger.error(f'Error in background initialization: {ex}')

    def initGraphics(self):
        self.buttonPortfolio.setStyleSheet("QPushButton {border-image: url(" +
                                           os.path.join(self.graphicsPath, 'portfolio.png').replace('\\', '/') +
                                           ") 0 15 0 15 stretch;}")
        self.buttonTrades.setStyleSheet("QPushButton {border-image: url(" +
                                        os.path.join(self.graphicsPath, 'trades.png').replace('\\', '/') +
                                        ") 0 15 0 15 stretch;}")
        self.buttonImport.setStyleSheet("QPushButton {border-image: url(" +
                                        os.path.join(self.graphicsPath, 'import.png').replace('\\', '/') +
                                        ") 0 15 0 15 stretch;}")
        self.buttonExport.setStyleSheet("QPushButton {border-image: url(" +
                                        os.path.join(self.graphicsPath, 'export.png').replace('\\', '/') +
                                        ") 0 15 0 15 stretch;}")
        self.buttonSettings.setStyleSheet("QPushButton {border-image: url(" +
                                          os.path.join(self.graphicsPath, 'settings.png').replace('\\', '/') +
                                          ") 0 15 0 15 stretch;}")


    def reinit(self):
        self.startRefresh.emit()
        # delete button styles before style update, otherwise qt will crash
        self.buttonPortfolio.setStyleSheet("")
        self.buttonTrades.setStyleSheet("")
        self.buttonImport.setStyleSheet("")
        self.buttonExport.setStyleSheet("")
        self.buttonSettings.setStyleSheet("")
        self.initStyle()
        self.initGraphics()
        self.endRefresh.emit()

    # load environment/ settings
    def initEnv(self):
        # handle paths
        self.appPath = application_path
        self.graphicsPath = getResourcePath('graphics')
        self.ImportPath = getResourcePath('Import')

        if self.username:
            dataFolderName = "Data_" + str(self.username)
        else:
            dataFolderName = "Data"

        # check if valid userDataPath has been provided as argument
        if self.userDataPath and os.path.isdir(self.userDataPath):
            # check access to data folder (does not work on windows)
            if not os.access(self.userDataPath, os.X_OK):
                print('provided datadir is not writable')
                sys.exit(0)
            # try to write to userDataPath
            self.dataPath = os.path.join(self.userDataPath, dataFolderName)
            try:
                os.stat(self.userDataPath)
                self.logger = logger.globalLogger.setPath(self.dataPath)
            except Exception:
                # try to create userDataPath
                try:
                    os.mkdir(self.dataPath)
                    self.logger = logger.globalLogger.setPath(self.dataPath)
                except Exception as ex:
                    print('provided datadir is not valid: ' + str(ex))
                    sys.exit(0)
        else:  # no datadir provided, try to find writable location
            # check access to app folder (does not work on windows)
            if os.access(self.appPath, os.X_OK):
                self.userDataPath = self.appPath
            else:
                print('user data can not be saved in koala dir: ' + str(self.appPath))
                from PyQt5.QtCore import QStandardPaths
                self.userDataPath = os.path.abspath(QStandardPaths.writableLocation(
                                                    QStandardPaths.AppDataLocation))
                print('user data will be saved in: ' + str(self.userDataPath))
                if not os.path.isdir(self.userDataPath):
                    try:
                        os.mkdir(self.userDataPath)
                    except Exception as ex:
                        print('error creating user data folder: ' + str(ex))
                        sys.exit(0)

            self.dataPath = os.path.join(self.userDataPath, dataFolderName)

            # check DataFolder
            try:
                os.stat(self.dataPath)
                self.logger = logger.globalLogger.setPath(self.dataPath)  # test logfile
            except:  # folder not found/ not writable
                try:  # try to create Data folder in app path
                    os.mkdir(self.dataPath)
                    self.logger = logger.globalLogger.setPath(self.dataPath)
                except:  # creation of data folder not possible
                    # create data folder in system appDataPath
                    print('user data can not be saved in dir: ' + str(self.appPath))
                    self.userDataPath = os.path.abspath(QStandardPaths.writableLocation(
                                                        QStandardPaths.AppDataLocation))
                    print('user data will be saved in: ' + str(self.userDataPath))
                    self.dataPath = os.path.join(self.userDataPath, dataFolderName)
                    try:
                        os.stat(self.dataPath)
                        self.logger = logger.globalLogger.setPath(self.dataPath)
                    except:  # folder not found/ not writable
                        try:  # try to create Data folder in app path (could be skipped in case of PermissionError)
                            if not os.path.isdir(self.userDataPath):
                                os.mkdir(self.userDataPath)
                            os.mkdir(self.dataPath)
                            self.logger = logger.globalLogger.setPath(self.dataPath)
                        except Exception as ex:
                            print('creating data folder failed: ' + str(ex))
                            sys.exit(0)

        # valid data path, create user data
        #self.logger = logger.globalLogger.setPath(self.dataPath)
        self.settings = settings.mySettings.setPath(self.dataPath)
        self.styleSheetHandler = style.StyleSheetHandler(self)
        self.styleSheetHandler.setPath(self.dataPath)
        self.exportTranslator = translator.ExportTranslator(dataPath=self.dataPath)
        style.myStyle = self.styleSheetHandler
        QTimer.singleShot(500, self.initDBComponents)

    def initDBComponents(self):
        """Initialize db components in background"""
        try:
            self.apiUserDatabase = apiImport.ApiUserDatabase(path=self.dataPath)
            self.apiDefaultDatabase = apiImport.ApiDefaultDatabase(path=self.ImportPath)

            self.logger.info('databases initialized')
        except Exception as ex:
            logger.globalLogger.error('error initializing db components: ' + str(ex))


    # setup window style
    def initWindow(self):
        self.logger.info('initializing window ...')
        #        self.setWindowFlags(qtcore.Qt.FramelessWindowHint)
        windowProps = self.settings.getWindowProperties()
        self.setWindowState(windowProps['state'])
        self.setGeometry(windowProps['geometry'])
        self.logger.info('window initialized')

    def initStyle(self):
        self.logger.info('initializing style ...')
        windowTitle = self.settings['window']['windowtitle']
        if self.username:
            windowTitle += ' ' + str(self.username)
        if self.latestVersion:
            windowTitle += ' ' + self.settings['general']['version'] + ' (latest: ' + str(self.latestVersion) + ')'
        else:
            windowTitle += ' ' + self.settings['general']['version']
        self.setWindowTitle(windowTitle)
        try:
            app_icon = QIcon()
            app_icon.addFile(os.path.join(self.appPath, 'KoalaIcon.png'), QSize(256, 256))
            self.setWindowIcon(app_icon)
        except Exception as ex:
            print(str(ex))
        try:
            from PyQt5.QtWidgets import QStyleFactory
            self.setStyle(QStyleFactory.create(self.settings['window']['windowstyle']))
        except Exception as ex:
            self.logger.error('window style is invalid: ' + str(ex))
            self.logger.info('using window style Fusion')
            self.setStyle(QStyleFactory.create('Fusion'))
        # load stylesheet
        self.styleSheetHandler.loadSheet(self.settings['window']['stylesheetname'])
        self.logger.info('style initialized')
        self.styleSheetChanged.connect(StyledCompleter.updateAllStyles)
        self.styleSheetChanged.emit(self.styleSheet())

    def initData(self):
        self.logger.info('initializing data ...')
        self.logList = loggerwidget.QLogModel()
        self.settingsModel = settingsPage.SettingsModel(self.settings)
        self.tradeList = ttable.QTradeTableModel(self.dataPath)
        self.coinList = ptable.QPortfolioTableModel(self.dataPath)

        self.logger.newLogMessage.connect(lambda status, statusType: self.logList.addString(status, statusType))
        self.tradeList.tradesAdded.connect(lambda tradeList: self.coinList.addTrades(tradeList))
        self.tradeList.tradesRemoved.connect(lambda tradeList: self.coinList.deleteTrades(tradeList))
        self.tradeList.histPricesUpdated.connect(self.coinList.histPricesChanged)
        self.tradeList.histPriceUpdateFinished.connect(self.coinList.histPriceUpdateFinished)
        self.settingsModel.displayCurrenciesChanged.connect(self.coinList.updateDisplayCurrencies)
        self.settingsModel.useWalletTaxFreeLimitYearsChanged.connect(self.coinList.taxYearWalletChanged)
        self.settingsModel.displayCurrenciesChanged.connect(self.tradeList.updateDisplayCurrencies)

        self.logger.info('data initialized')


    # setup layout
    def layoutUI(self):
        self.logger.info('initializing gui layout ...')
        self.mainLayout = QHBoxLayout(self)
        self.setContentsMargins(0, 0, 0, 0)

        # sidebar with buttons for pagecontrol
        from PyQt5.QtWidgets import QFrame, QPushButton
        self.sidebarFrame = QFrame(self)
        self.sidebarFrame.setFrameShape(QFrame.StyledPanel)
        self.sidebarFrame.setFrameShadow(QFrame.Raised)
        self.sidebarFrame.setFixedWidth(120)

        # buttons sidebar
        self.buttonPortfolio = QPushButton("", self.sidebarFrame)
        self.buttonTrades = QPushButton("", self.sidebarFrame)
        self.buttonImport = QPushButton("", self.sidebarFrame)
        self.buttonExport = QPushButton("", self.sidebarFrame)
        self.buttonSettings = QPushButton("", self.sidebarFrame)

        self.initGraphics()

        self.buttonPortfolio.setToolTip('portfolio')
        self.buttonTrades.setToolTip('trades')
        self.buttonImport.setToolTip('import')
        self.buttonExport.setToolTip('export')
        self.buttonSettings.setToolTip('settings')

        self.buttonPortfolio.clicked.connect(lambda: self.showFrame(self.PORTFOLIOPAGEINDEX))
        self.buttonTrades.clicked.connect(lambda: self.showFrame(self.TRADESPAGEINDEX))
        self.buttonImport.clicked.connect(lambda: self.showFrame(self.IMPORTPAGEINDEX))
        self.buttonExport.clicked.connect(lambda: self.showFrame(self.EXPORTPAGEINDEX))
        self.buttonSettings.clicked.connect(lambda: self.showFrame(self.SETTINGSPAGEINDEX))
        buttonHeight = 120
        self.sidebarButtons = [self.buttonPortfolio, self.buttonTrades, self.buttonImport, self.buttonExport, self.buttonSettings]
        self.sidebarLayout = QVBoxLayout(self.sidebarFrame)
        for button in (self.sidebarButtons):
            button.setFixedHeight(buttonHeight)
            #            button.setCheckable(True)
            self.sidebarLayout.addWidget(button)

        self.sidebarLayout.addStretch()

        # %% statusbar for displaying status and progress of ongoing actions
        self.statusbar = controls.StatusBar(self, height=80, dataPath=self.dataPath)
        self.statusbar.setModel(self.logList)

        # %%  stacked Layout for content frames
        self.portfolioPage = PortfolioPage(parent=self, controller=self)
        self.tradesPage = TradesPage(parent=self, controller=self)
        self.importPage = ImportPage(parent=self, controller=self)
        self.exportPage = exportpage.ExportPage(parent=self, controller=self)
        self.settingsModel.useWalletTaxFreeLimitYearsChanged.connect(self.exportPage.taxYearWalletChanged)
        self.settingsPage = pages.SettingsPage(parent=self, controller=self)
        self.pages = [self.portfolioPage, self.tradesPage, self.importPage, self.exportPage, self.settingsPage]
        from PyQt5.QtWidgets import QStackedLayout
        self.stackedContentLayout = QStackedLayout()
        for page in self.pages:
            # stacked content layout
            self.stackedContentLayout.addWidget(page)

        # put frames in vertical layout
        self.verticalFrameLayout = QVBoxLayout()
        self.verticalFrameLayout.setSpacing(0)
        self.verticalFrameLayout.addLayout(self.stackedContentLayout)
        self.verticalFrameLayout.addWidget(self.statusbar)

        # put frames in horizontal layout
        self.horizontalFrameLayout = QHBoxLayout()
        self.horizontalFrameLayout.addWidget(self.sidebarFrame)
        self.horizontalFrameLayout.addLayout(self.verticalFrameLayout)

        # put subLayouts in main layout
        self.mainLayout.addLayout(self.horizontalFrameLayout)

        self.logger.info('gui layout initialized')

    def startThreads(self):
        self.logger.info('starting threads ...')
        self.priceThread = threads.UpdatePriceThread(self.coinList, self.tradeList)
        self.priceThread.coinPricesLoaded.connect(lambda prices, coins: self.coinList.setPrices(prices, coins))
        self.priceThread.coinIconsLoaded.connect(lambda icons, coins: self.coinList.setIcons(icons, coins))
        self.priceThread.coinPriceChartsLoaded.connect(lambda priceChartData: self.coinList.setPriceChartData(priceChartData))
        self.priceThread.historicalPricesLoaded.connect(lambda prices, tradesLeft:
                                                        self.tradeList.setHistPrices(prices, tradesLeft))
        self.priceThread.start()
        self.logger.info('threads started')

    def showFrame(self, pageIndex):
        """Show a frame for the given page index"""
        frame = self.pages[pageIndex]
        if hasattr(frame, 'refresh'):
            frame.refresh()
        self.stackedContentLayout.setCurrentIndex(pageIndex)
        # reset color of all buttons in sidebar
        # for buttonIndex in range(len(self.sidebarButtons)):
        #     button = self.sidebarButtons[buttonIndex]
        #     if buttonIndex == pageIndex:
        #         # do something with selected button
        #         #                button.config(bg=controls.changeColorBrightness(self.sidebar['bg'], 40))
        #         pass
        #     else:
        #         # do something with not selected buttons
        #         #                button.config(bg = self.sidebar['bg'])
        #         pass

    def closeEvent(self, event):
        # save window props
        geometry = self.geometry()
        state = self.windowState()
        self.settings.setWindowProperties(geometry, state)
        # save gui props
        self.settings.setGuiSettings(self.tradesPage.getGuiProps())
        self.settings.setGuiSettings(self.portfolioPage.getGuiProps())
        self.settings.saveSettings()
        event.accept()

    def printStatus(self, status, statusType='i'):
        self.logList.addString(status, statusType)

    # get TradeList
    def getTradeList(self):
        return self.tradeList

    # get CoinList
    def getCoinList(self):
        return self.coinList


def main():
    args = arguments.parse_arguments()
    coreLogger.globalLogger = logger.globalLogger
    coreSettings.mySettings = settings.mySettings
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    app = QApplication(argv)

    # Create and show splash screen with better styling
    try:
        splash_pixmap = QPixmap(getResourcePath('graphics/KoalaIcon.ico'))
        if splash_pixmap.isNull():
            # Fallback to PNG if ICO is not found
            splash_pixmap = QPixmap(getResourcePath('graphics/KoalaIcon.png'))
    except:
        # Create a simple colored splash if no icon is found
        splash_pixmap = QPixmap(300, 200)
        splash_pixmap.fill(Qt.darkGray)

    splash = CustomSplashScreen(splash_pixmap)
    splash.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
    splash.showMessage("Loading Koalafolio...",
                      Qt.AlignBottom | Qt.AlignCenter,
                      Qt.white)
    splash.show()
    app.processEvents()  # Process events to make sure splash is displayed immediately

    try:
        os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
        app_icon = QIcon()
        app_icon.addFile(getResourcePath('graphics/KoalaIcon.png'), QSize(256, 256))
        app.setWindowIcon(app_icon)
        app.setApplicationName('koalafolio')
    except Exception as ex:
        print(str(ex))

    # Initialize main window with splash reference
    window = PortfolioApp(userDataPath=args.datadir, username=args.username, splash=splash)

    # Close splash when main window is ready
    splash.finish(window)

    exit(app.exec_())

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created on 29.11.2020

@author: Martin
"""

from pandas import DataFrame
from datetime import datetime
import ccxt
from time import sleep
from typing import List, Optional, Dict, Any, Set
import koalafolio.gui.helper.QLogger as logger

localLogger = logger.globalLogger

# static class for static data
class ExchangesStatic:
    SUPPORTED_EXCHANGES = [exchange_id for exchange_id in ccxt.exchanges if getattr(ccxt, exchange_id)().has.get('fetchMyTrades', False)]

    # convert ccxtTrades to Dataframe
    @staticmethod
    def tradesToDataframe(ccxtTrades):
        # timestamp,  location,   pair,   trade_type, amount, rate,   fee,    fee_currency,   link
        trades = []
        for ccxtTrade in ccxtTrades:
            trade = {'timestamp': datetime.fromtimestamp(ccxtTrade.timestamp),
                     'location': str(ccxtTrade.location),
                     'pair': str(ccxtTrade.pair),
                     'trade_type': str(ccxtTrade.trade_type),
                     'amount': float(ccxtTrade.amount),
                     'rate': float(ccxtTrade.rate),
                     'fee': float(ccxtTrade.fee),
                     'fee_currency': str(ccxtTrade.fee_currency.symbol),
                     'link': str(ccxtTrade.link)}
            trades.append(trade)
        return DataFrame(trades)

    # get trade history from exchange
    @staticmethod
    def getTradeHistoryCcxt(apiname, key, secret, start, end):
        # Handle different secret encoding requirements for different exchanges
        api_secret = secret
        if isinstance(secret, str) and apiname.lower() != 'binance':
            api_secret = secret.encode()

        api = ccxtExchange(apiname, api_key=key, api_secret=api_secret)
        iskeyValid, checkKeyMsg = api.validate_api_credentials()
        if iskeyValid:
            trades = api.query_trade_history(start_ts=start, end_ts=end)
            tradesDF = ExchangesStatic.tradesToDataframe(trades)
            return tradesDF
        localLogger.warning("api key is invalid for " + str(api.__class__.__name__) + ": " + checkKeyMsg)
        return DataFrame()


class Trade:
    """Simple class to represent a trade in the format expected by tradesToDataframe"""
    def __init__(self, timestamp, location, pair, trade_type, amount, rate, fee, fee_currency, link=""):
        self.timestamp = timestamp
        self.location = location
        self.pair = pair
        self.trade_type = trade_type
        self.amount = amount
        self.rate = rate
        self.fee = fee
        self.fee_currency = type('obj', (object,), {'symbol': fee_currency})
        self.link = link


class ccxtExchange:
    """CCXT Exchange wrapper for koalafolio"""
    
    SUPPORTED_EXCHANGES = ExchangesStatic.SUPPORTED_EXCHANGES
    
    def __init__(self, exchange_name, api_key=None, api_secret=None):
        """
        Initialize the CCXT exchange wrapper
        
        Args:
            exchange_name (str): Name of the exchange (must be one of the supported exchanges)
            api_key (str): API key for the exchange
            api_secret (str): API secret for the exchange
        """
        if exchange_name.lower() not in self.SUPPORTED_EXCHANGES:
            raise ValueError(f"Exchange {exchange_name} not supported. Supported exchanges: {', '.join(self.SUPPORTED_EXCHANGES)}")
        
        self.exchange_name = exchange_name.lower()
        self.api_key = api_key
        self.api_secret = api_secret
        
        # Initialize the CCXT exchange
        exchange_class = getattr(ccxt, self.exchange_name)
        self.exchange = exchange_class({
            'apiKey': self.api_key,
            'secret': self.api_secret,
            'enableRateLimit': False,  # Respect exchange rate limits
        })
    
    def validate_api_credentials(self):
        """
        Validate the API credentials by attempting to fetch account balance
        
        Returns:
            tuple: (is_valid, message) where is_valid is a boolean and message is a string
        """
        try:
            self.exchange.fetch_balance()
            return True, "API credentials are valid"
        except Exception as e:
            return False, str(e)
    
    def _is_symbol_required_for_trades(self):
        """
        Check if the exchange requires a symbol for fetching trades
        
        Returns:
            bool: True if symbol is required, False otherwise
        """
        try:
            # Check if the exchange has the features property
            if hasattr(self.exchange, 'describe') and callable(self.exchange.describe):
                description = self.exchange.describe()
                
                # Navigate through the features structure
                if 'features' in description:
                    features = description['features']
                    
                    # Check spot trading features
                    if 'spot' in features and features['spot'] is not None:
                        spot_features = features['spot']
                        
                        # Check fetchMyTrades feature
                        if 'fetchMyTrades' in spot_features and spot_features['fetchMyTrades'] is not None:
                            fetch_trades_features = spot_features['fetchMyTrades']
                            
                            # Check if symbolRequired is specified
                            if 'symbolRequired' in fetch_trades_features:
                                return fetch_trades_features['symbolRequired']
            
            # If we can't determine from features, use a conservative approach
            # Check if the exchange has implemented fetchMyTrades without a symbol
            try:
                # Try to call fetchMyTrades with no symbol and a small limit
                # This is just a test call to see if it works
                self.exchange.fetch_my_trades(symbol=None, limit=1)
                return False  # If it works, symbol is not required
            except Exception as e:
                error_msg = str(e).lower()
                # Check for common error messages indicating symbol is required
                if 'symbol' in error_msg and ('required' in error_msg or 'must be' in error_msg):
                    return True
            
            # Default to requiring symbol for safety
            return True
            
        except Exception as e:
            localLogger.warning(f"Error checking if symbol is required: {str(e)}")
            return True  # Default to requiring symbol for safety
    
    def _get_user_assets(self) -> Set[str]:
        """
        Get the assets that the user has traded with
        
        Returns:
            set: Set of asset symbols
        """
        user_assets = set()
        
        # Try to get deposit history
        try:
            deposits = self.exchange.fetch_deposits()
            for deposit in deposits:
                if 'currency' in deposit:
                    user_assets.add(deposit['currency'])
        except Exception as e:
            localLogger.warning(f"Could not fetch deposit history: {str(e)}")
        
        # Try to get withdrawal history
        try:
            withdrawals = self.exchange.fetch_withdrawals()
            for withdrawal in withdrawals:
                if 'currency' in withdrawal:
                    user_assets.add(withdrawal['currency'])
        except Exception as e:
            localLogger.warning(f"Could not fetch withdrawal history: {str(e)}")
            
        # checking current balances
        try:
            balances = self.exchange.fetch_balance()
            if balances and 'total' in balances:
                for currency, amount in balances['total'].items():
                    if amount > 0:
                        user_assets.add(currency)
        except Exception as e:
            localLogger.warning(f"Could not fetch balance: {str(e)}")
        
        # add common assets
        user_assets.update({'BTC', 'ETH', 'BNB', 'USD', 'EUR', 'USDT', 'USDC'})
        
        return user_assets
    
    def _get_symbols_to_check(self, user_assets: Set[str]) -> List[str]:
        """
        Get the symbols to check for trades based on user assets
        
        Args:
            user_assets: Set of asset symbols the user has
            
        Returns:
            list: List of symbol strings to check
        """
        symbols_to_check = []
        
        try:
            # Only check symbols that contain the user's assets
            for symbol in self.exchange.markets:
                try:
                    market = self.exchange.market(symbol)
                    base, quote = market['base'], market['quote']
                    if base in user_assets and quote in user_assets:
                        symbols_to_check.append(symbol)
                except Exception:
                    continue
                    
        except Exception as e:
            localLogger.warning(f"Error determining symbols: {str(e)}")
            # Fallback to a minimal set of common pairs
            common_pairs = ['BTC/USDT', 'ETH/USDT', 'BNB/USDT', 'ETH/BTC']
            for pair in common_pairs:
                if pair in self.exchange.markets:
                    symbols_to_check.append(pair)
        
        return symbols_to_check
    
    def _fetch_trades_for_symbol(self, symbol: str, since: Optional[int] = None, limit: int = 50) -> List[Dict[str, Any]]:
        """
        Fetch trades for a specific symbol with rate limiting and retry logic
        
        Args:
            symbol: Trading pair symbol
            since: Timestamp in milliseconds to fetch trades from
            limit: Maximum number of trades to fetch per request
            
        Returns:
            list: List of trade dictionaries
        """
        all_trades = []
        
        try:
            # Initial request
            params = {}
            symbol_trades = self.exchange.fetch_my_trades(symbol=symbol, since=since, limit=limit, params=params)
            
            if symbol_trades:
                all_trades.extend(symbol_trades)
                
                # Some exchanges support pagination
                if len(symbol_trades) == limit:
                    # Continue fetching if we got exactly 'limit' trades (likely more available)
                    last_trade = symbol_trades[-1]
                    
                    # Different exchanges have different pagination mechanisms
                    if self.exchange_name == 'binance':
                        # Binance uses fromId for pagination
                        while len(symbol_trades) == limit:
                            # Add delay between requests to avoid rate limits
                            sleep(0.3)
                            
                            # Get the ID of the last trade
                            last_id = last_trade['id']
                            params = {'fromId': last_id}
                            
                            try:
                                symbol_trades = self.exchange.fetch_my_trades(symbol=symbol, limit=limit, params=params)
                                if symbol_trades:
                                    all_trades.extend(symbol_trades)
                                    last_trade = symbol_trades[-1]
                                else:
                                    break
                            except Exception as e:
                                localLogger.warning(f"Error during pagination for {symbol}: {str(e)}")
                                break
                    else:
                        # For other exchanges, use timestamp-based pagination
                        while len(symbol_trades) == limit:
                            # Add delay between requests to avoid rate limits
                            if self.exchange_name in ['coinbase', 'coinbasepro', 'gemini']:
                                sleep(0.5)
                            else:
                                sleep(0.2)
                            
                            # Get the timestamp of the last trade + 1ms
                            next_since = last_trade['timestamp'] + 1
                            
                            try:
                                symbol_trades = self.exchange.fetch_my_trades(symbol=symbol, since=next_since, limit=limit)
                                if symbol_trades:
                                    all_trades.extend(symbol_trades)
                                    last_trade = symbol_trades[-1]
                                else:
                                    break
                            except Exception as e:
                                localLogger.warning(f"Error during pagination for {symbol}: {str(e)}")
                                break
            
            # Add delay between requests to avoid rate limits
            if self.exchange_name == 'binance':
                sleep(0.3)  # 300ms delay between requests for Binance
            elif self.exchange_name in ['coinbase', 'coinbasepro', 'gemini']:
                sleep(0.5)  # 500ms delay for these exchanges
            else:
                sleep(0.2)  # 200ms default delay for other exchanges
                
            return all_trades
            
        except Exception as e:
            if 'rate limit' in str(e).lower():
                localLogger.warning(f"Rate limit hit for {symbol}, waiting 30 seconds before continuing")
                sleep(30)  # Wait 30 seconds if we hit rate limit
                try:
                    # Try again after waiting
                    return self._fetch_trades_for_symbol(symbol, since, limit)
                except Exception as retry_e:
                    localLogger.warning(f"Error fetching trades for {symbol} after retry: {str(retry_e)}")
            else:
                localLogger.warning(f"Error fetching trades for {symbol}: {str(e)}")
            
            return all_trades
    
    def _convert_raw_trade_to_trade_object(self, raw_trade: Dict[str, Any]) -> Optional[Trade]:
        """
        Convert a raw trade dictionary from CCXT to a Trade object
        
        Args:
            raw_trade: Dictionary containing trade data from CCXT
            
        Returns:
            Trade: Trade object or None if conversion fails
        """
        try:
            # Extract trade information
            timestamp = raw_trade['timestamp'] / 1000  # Convert from ms to seconds
            location = self.exchange_name
            pair = raw_trade['symbol']
            
            # Determine trade type (buy/sell)
            trade_type = raw_trade['side']
            
            # Extract amount and rate
            amount = raw_trade['amount']
            rate = raw_trade['price']
            
            # Extract fee information
            fee = 0
            fee_currency = ''
            
            if raw_trade['fee'] is not None:
                fee = raw_trade['fee'].get('cost', 0)
                fee_currency = raw_trade['fee'].get('currency', '')
            
            # Create link (trade ID or empty string)
            link = raw_trade.get('id', "")
            
            # Create Trade object
            return Trade(timestamp, location, pair, trade_type, amount, rate, fee, fee_currency, link)
            
        except Exception as e:
            localLogger.warning(f"Error processing trade: {str(e)}, trade data: {raw_trade}")
            return None
    
    def query_trade_history(self, start_ts=None, end_ts=None, limit_per_request=50):
        """
        Query trade history from the exchange with pagination
        
        Args:
            start_ts (int): Start timestamp in seconds since epoch
            end_ts (int): End timestamp in seconds since epoch
            limit_per_request (int): Maximum number of trades to fetch per request
            
        Returns:
            list: List of Trade objects
        """
        try:
            # Convert timestamps to milliseconds if provided
            since = int(start_ts * 1000) if start_ts else None
            
            # Fetch trades from the exchange
            raw_trades = []
            
            # Some exchanges require fetching trades by symbol
            if not self.exchange.has['fetchMyTrades']:
                localLogger.warning(f"Exchange {self.exchange_name} does not support fetching trades")
                return []
            
            # Get all markets first
            self.exchange.load_markets()
            
            # Check if symbol is required for fetching trades
            symbol_required = self._is_symbol_required_for_trades()
            localLogger.info(f"Symbol required for {self.exchange_name}: {symbol_required}")
            
            # For exchanges that don't require a symbol
            if not symbol_required:
                params = {}
                has_more = True
                current_since = since
                
                while has_more:
                    batch = self.exchange.fetch_my_trades(symbol=None, since=current_since, limit=limit_per_request, params=params)
                    if batch:
                        raw_trades.extend(batch)
                        
                        # If we got exactly the limit, there might be more
                        if len(batch) == limit_per_request:
                            # Update since for next request
                            current_since = batch[-1]['timestamp'] + 1
                        else:
                            has_more = False
                    else:
                        has_more = False
                        
                    # Add delay between requests
                    sleep(0.3)
                
                 # Kraken fetch_my_trades only returns trades since 2022, if start is before 2022 add warning for kraken
                if self.exchange_name == 'kraken' and start_ts and start_ts < 1640995200:  # Before 2022
                    localLogger.warning("Kraken Trade History older then 2022 not supported by Kraken API. Export history older then 2022 as .csv from Kraken Website")
  
                
            else:
                # For exchanges that require fetching trades by symbol
                # Get user assets to determine which symbols to check
                user_assets = self._get_user_assets()
                
                # Get symbols to check based on user assets
                localLogger.info(f"Only check symbols for user assets: {user_assets}")
                symbols_to_check = self._get_symbols_to_check(user_assets)
                
                # Fetch trades for each symbol with pagination
                localLogger.info(f"Number of symbols to check: {len(symbols_to_check)}")
                for symbol in symbols_to_check:
                    symbol_trades = self._fetch_trades_for_symbol(symbol, since, limit_per_request)
                    raw_trades.extend(symbol_trades)
            
            # Filter trades by end timestamp if provided
            if end_ts:
                raw_trades = [t for t in raw_trades if t['timestamp'] / 1000 <= end_ts]
            
            # Convert raw trades to Trade objects
            trades = []
            for raw_trade in raw_trades:
                trade = self._convert_raw_trade_to_trade_object(raw_trade)
                if trade:
                    trades.append(trade)
            
            localLogger.info(f"Found {len(trades)} trades for {self.exchange_name}")
            return trades
            
        except Exception as e:
            localLogger.error(f"Error querying trade history from {self.exchange_name}: {str(e)}")
            return []


# -*- coding: utf-8 -*-
"""
Created on Thu Oct  4 15:15:19 2018

@author: Martin
"""

# todo: check why default section is written to file ?!


import koalafolio.gui.helper.QLogger as logger
from PyQt5.QtWidgets import QLineEdit, QStyledItemDelegate
from PyQt5.QtCore import QAbstractItemModel, QSize, QVariant, Qt, pyqtSignal
import koalafolio.gui.widgets.ScrollableTable as sTable

localLogger = logger.globalLogger

class SettingsModelItem():
    def __init__(self, key, parentkey, parent):
        self.key = key
        self.parentKey = parentkey
        self.parent = parent

    def isEqual(self, other):
        return self.key == other.key and self.parentKey == other.parentKey


# settings model
class SettingsModel(QAbstractItemModel):
    displayCurrenciesChanged = pyqtSignal([list])
    useWalletTaxFreeLimitYearsChanged = pyqtSignal()

    def __init__(self, settings, *args, **kwargs):
        super(SettingsModel, self).__init__(*args, **kwargs)
        self.settings = settings
        self.sections = list(self.settings)[1:]
        self.items = []
        self.header = ['setting', 'value', 'description']

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole:
            return self.header[section]
        return QVariant()

    def parent(self, index):
        return self.getParentIndexFromIndex(index)

    def rowCount(self, parent):
        if not parent.isValid():
            return len(self.sections)
        else:
            # check if it is a parent or a child item
            if self.getParentIndexFromIndex(parent).isValid():  # only a child item has a valid parent
                return 0  # return no childs
            else:  # parent item, return number of childs
                return len(self.settings[self.getItemFromIndex(parent).key])

    def columnCount(self, parent):
        return 3

    def index(self, row, column, parent):
        if parent.isValid():
            parentkey = self.getItemFromIndex(parent).key
            subsettings = self.settings[parentkey]
            try:
                key = list(subsettings)[row]
            except IndexError:
                key = ""
        else:
            parentkey = ""
            try:
                key = self.sections[row]
            except IndexError:
                key = ""
        newItem = SettingsModelItem(key, parentkey, parent)
        itemId = -1
        itemFound = False
        for item in self.items:
            itemId += 1
            if item.isEqual(newItem):
                itemFound = True
                break
        if not itemFound:
            itemId = len(self.items)
            self.items.append(newItem)
        index = self.createIndex(row, column, itemId)
        return index

    def data(self, index, role):
        if role == Qt.DisplayRole:
            parentIndex = self.getParentIndexFromIndex(index)
            item = self.getItemFromIndex(index)
            if parentIndex.isValid():
                parent = self.getItemFromIndex(parentIndex)
                if index.column() == 0:
                    return item.key
                if index.column() == 1:
                    return self.settings[parent.key][item.key]
                if index.column() == 2:
                    try:
                        return self.settings.descriptions[parent.key][item.key]
                    except KeyError:
                        return 'depricated/ unknowen'
            else:
                return self.sections[index.row()]
        return QVariant()

    def setData(self, index, value, role):
        if role == Qt.EditRole:
            try:
                item = self.getItemFromIndex(index)
                self.settings[item.parentKey][item.key] = value
                self.dataChanged.emit(index, index)
                self.itemChanged(item)
                return True
            except:
                return False
        return False

    def itemChanged(self, item):
        if item.parentKey.lower() == 'currency' and item.key.lower() == 'defaultdisplaycurrencies':
            self.displayCurrenciesChanged.emit(self.settings.displayCurrencies())
        if item.parentKey.lower() == 'tax' and item.key.lower() == 'usewallettaxfreelimityears':
            self.useWalletTaxFreeLimitYearsChanged.emit()

    def resetDefault(self):
        self.beginResetModel()
        oldCur = self.settings.displayCurrencies()
        self.settings.resetDefault()
        if oldCur != self.settings.displayCurrencies():
            self.displayCurrenciesChanged.emit(self.settings.displayCurrencies())
        self.endResetModel()

    def restoreSettings(self):
        oldCur = self.settings.displayCurrencies()
        self.settings.readSettings()
        if oldCur != self.settings.displayCurrencies():
            self.displayCurrenciesChanged.emit(self.settings.displayCurrencies())

    def flags(self, index):
        column = index.column()
        parent = self.getParentIndexFromIndex(index)
        if parent.isValid():
            if column == 1:
                return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemNeverHasChildren
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemNeverHasChildren
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def getParentIndexFromIndex(self, index):
        return self.items[index.internalId()].parent

    def getItemFromIndex(self, index):
        return self.items[index.internalId()]

    def getParentFromIndex(self, index):
        parentIndex = self.getParentIndexFromIndex(index)
        if parentIndex.isValid():
            return self.getItemFromIndex(parentIndex)
        else:
            return None


class SettingsDelegate(QStyledItemDelegate):
    def __init__(self, *args, **kwargs):
        super(SettingsDelegate, self).__init__(*args, **kwargs)

    def createEditor(self, parent, option, index):
        if int(index.flags()) & Qt.ItemIsEditable:
            return QLineEdit(parent)
        return 0

    def setEditorData(self, editor, index):
        editor.setText(index.data())

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.text(), Qt.EditRole)

    def sizeHint(self, option, index):
        return QSize()


class SettingsTreeView(sTable.QScrollableTreeView):
    def __init__(self, *args, **kwargs):
        super(SettingsTreeView, self).__init__(*args, **kwargs)

        self.setItemDelegateForColumn(1, SettingsDelegate())



# -*- coding: utf-8 -*-
"""
Created on Sun Sep  15 15:15:19 2019

@author: Martin
"""
from koalafolio.web.exchanges import ExchangesStatic
from koalafolio.web.chaindata import ChaindataStatic
from koalafolio.gui.helper.QLogger import globalLogger
from pandas import DataFrame
from Cryptodome.Cipher import AES
from os import remove as os_remove
from os.path import exists as os_exists, join as os_join, isfile
from hashlib import sha256
from json import dumps as json_dumps, loads as json_loads
from Cryptodome.Cipher._mode_eax import EaxMode

localLogger = globalLogger


# static class for static data
class ApiImportStatic:
    SUPPORTED_APIS = ExchangesStatic.SUPPORTED_EXCHANGES + ChaindataStatic.SUPPORTED_CHAINS
    @staticmethod
    def getApiHistory(apiname, apitype, start, end, apikey=None, secret=None, address=None):
        if apiname not in ApiImportStatic.SUPPORTED_APIS:
            raise KeyError("invalid api name")
        else:
            localLogger.info("requesting data from " + str(apiname))
            try:
                if apitype == "exchange":
                    return ExchangesStatic.getTradeHistoryCcxt(apiname, apikey, secret, start, end)
                elif apitype == "chaindata":
                    return ChaindataStatic.getBlockdaemonRewardsForAddress(apiname, apikey, address, start, end)
                else:
                    raise ValueError("invalid type in getApiHistory: " + str(apitype))
            except Exception as ex:
                localLogger.warning("could not load data from api (" + str(apiname) + "): " + str(ex))
                return DataFrame()

    # get dict of ApiModels
    @staticmethod
    def getApiModels():
        apiModels = {}
        for apiname in ExchangesStatic.SUPPORTED_EXCHANGES:
            apiModels[apiname] = ApiModel(
                name=str(apiname),
                apitype="exchange",
                note="get data from " + str(apiname) + ". "
            )
        for apiname in ChaindataStatic.SUPPORTED_CHAINS:  # apitype is blockdaemon chaindata
            apiModels[apiname] = ApiModel(
                name=str(apiname),
                apitype="chaindata",
                note="Koala provides default api key for blockdaemon api.\n" \
                     "However it is recommended to create your own key.\n" \
                     "Key can be created at https://app.blockdaemon.com (Login->workspace->API Suite->Connect.\n" \
                     "More Infos can be found in their documentation: https://docs.blockdaemon.com/"
            )
        return apiModels


# %% Trade
class ApiModel:
    def __init__(self, name='', apitype='', note=''):
        self.apiName = name
        self.apiType = apitype
        self.apiNote = note

    def toDict(self):
        return {
            "apiName": self.apiName,
            "apiType": self.apiType,
            "apiNote": self.apiNote
        }


class BaseDatabase:
    def __init__(self, dbname: str, path: str=None, loggingenabled: bool=False):

        self.databaseFound = False
        self.databaseLocked = True
        self.data = None
        self.filepath = None
        self.filename = dbname
        self.loggingEnabled = loggingenabled
        if path:
            self.setPath(path)

    def logWarning(self, msg: str):
        if self.loggingEnabled:
            localLogger.warning(msg)

    def logInfo(self, msg: str):
        if self.loggingEnabled:
            localLogger.info(msg)

    def logError(self, msg: str):
        if self.loggingEnabled:
            localLogger.error(msg)

    def setPath(self, path: str):
        # check path
        if os_exists(path):
            # set path
            self.filepath = os_join(path, self.filename).replace('\\', '/')
            if isfile(self.filepath):
                self.databaseFound = True
            else:
                self.databaseFound = False

    def encryptData(self, pw: str, data: dict) -> tuple[EaxMode, bytes, bytes]:
        dbkey = sha256(pw.encode()).hexdigest()
        cipher = AES.new(dbkey[:32].encode('utf-8'), AES.MODE_EAX)
        ciphertext, tag = cipher.encrypt_and_digest(json_dumps(data).encode("utf-8"))
        return cipher, ciphertext, tag

    def decryptData(self, pw: str, ciphertext: bytes, tag: bytes, nonce: bytes) -> dict:
        key = sha256(pw.encode()).hexdigest()
        cipher = AES.new(key[:32].encode('utf-8'), AES.MODE_EAX, nonce)
        return json_loads(cipher.decrypt_and_verify(ciphertext, tag))

    def checkPw(self, pw: str) -> bool:
        with open(self.filepath, "rb") as file_in:
            nonce, tag, ciphertext = [file_in.read(x) for x in (16, 16, -1)]
        try:
            self.decryptData(pw, ciphertext, tag, nonce)
            return True
        except ValueError:
            return False

    def readDatabase(self, pw: str) -> dict:
        with open(self.filepath, "rb") as file_in:
            nonce, tag, ciphertext = [file_in.read(x) for x in (16, 16, -1)]
        try:
            self.data = self.decryptData(pw, ciphertext, tag, nonce)
            self.databaseLocked = False
            self.logInfo("api database loaded")
            return self.data
        except ValueError:
            self.logWarning("decrypting api database failed, check password")
            return None

    def saveDatabase(self, pw: str):
        if self.checkPw(pw):
            cipher, ciphertext, tag = self.encryptData(pw, self.data)
            with open(self.filepath, "wb") as file_out:
                [file_out.write(x) for x in (cipher.nonce, tag, ciphertext)]
        else:
            self.logWarning("invalid password")

    def addDBEntry(self, pw: str, apiname: str, newData: dict):
        if not self.checkPw(pw):
            self.logWarning("invalid password")
            return
        if self.data is not None:
            self.data[apiname] = newData
            self.saveDatabase(pw)
        else:
            self.logWarning("api database need be be unlocked before it can be changed")

    def deleteDBEntry(self, pw: str, apiname: str):
        if not self.checkPw(pw):
            self.logWarning("invalid password")
            return
        if self.data is not None:
            if apiname in self.data:
                self.data.pop(apiname)
                self.saveDatabase(pw)
        else:
            self.logWarning("api database need be be unlocked before it can be changed")

    def createDatabase(self, pw: str):
        if self.databaseFound:
            self.logWarning("database already created, please delete the old database to create a new one.")
        else:
            self.data = {}
            cipher, ciphertext, tag = self.encryptData(pw, self.data)
            with open(self.filepath, "wb") as file_out:
                [file_out.write(x) for x in (cipher.nonce, tag, ciphertext)]
            self.databaseFound = True
            self.databaseLocked = False
            self.logInfo("api database created")

    def lockDatabase(self):
        self.data = None
        self.databaseLocked = True

    def deleteDatabase(self):
        self.lockDatabase()
        os_remove(self.filepath)
        self.databaseFound = 0


# database for user data. Will be created on user request
class ApiUserDatabase(BaseDatabase):
    def __init__(self, path: str=None):
        super(ApiUserDatabase, self).__init__(dbname='key.db', path=path, loggingenabled=True)


# database for default data. Will be uploaded to github and stores default data like default api keys
# api keys are created from koala specific account, users can create their own keys later
# encrypt data so it can not be parsed so easily from repo.
# all data in this default db is potentially exposed so no critical data will be stored here
class ApiDefaultDatabase(BaseDatabase):
    def __init__(self, path: str=None):
        if path:
            super(ApiDefaultDatabase, self).__init__(dbname='defaultApiData.db', path=path, loggingenabled=False)
            self.pw = "koala"

            # create
            # self.createDatabase(self.pw)
            # newKey = "some_key"
            # self.addDBEntry(self.pw, apiname="blockdaemon", newData=newKey)

            # read database
            self.readDatabase(self.pw)
            # print("defaultDBData: " + str(self.data))


